module SSG.HTML

import public SSG.HTML.ElementTypes

import Data.String

import Structures.Dependent.DList

export infix 8 =$

public export
data Attribute : Type where
  -- TODO: Add some structure here, maybe make `=$` a smart constructor
  (=$) : (attribute, value : String) -> Attribute

namespace Attribute
  export
  toString : Attribute -> String
  toString (attribute =$ value) =
    let value_cs = unpack value
    in if any (== '\'') value_cs
      then "\"\{value}\""
      else if any (== '"') value_cs || any (== ' ') value_cs
        then "'\{value}'"
        else value

public export
data Tag : String -> Type where
  -- TODO: Find a way to move this out of tag
  ||| Raw text, represented as a tag for compatibility
  Raw : (content : String)
    -> Tag ""
  ||| A void type tag, has attributes, no contents
  Void : (type : String) -> {auto void_prf : IsVoid type}
    -> (attributes : List Attribute)
    -> Tag type
  ||| A normal type tag, has attributes, and contents
  Normal : (type : String) -> {auto normal_prf : IsNormal type}
    -> (attributes : List Attribute)
    -- TODO: Erase content_types
    -> {content_types : _} -> (contents : DList _ Tag content_types)
    -> Tag type
  ||| A raw text or an escapable raw text element
  RawText : (type : String)
    -> {auto raw_prf : Either (IsRawText type) (IsEscapableRawText type)}
    -> (attributes : List Attribute)
    -> (content : String)
    -> Tag type

namespace Tag
  viewIndented : (indent_level : Nat) -> Tag type -> String
  viewIndented indent_level (Raw content) =
    replicate (indent_level * 2) ' ' ++ content
  viewIndented indent_level (Void type attributes) =
    let indent = replicate (indent_level * 2) ' '
    in if length attributes > 0
      then
        let attrs = joinBy " " $ map toString attributes
        in "<\{type} \{attrs} />"
      else "<\{type} />"
  viewIndented indent_level (Normal type attributes contents) =
    ?viewIndented_rhs_2
  viewIndented indent_level (RawText type attributes content) =
    ?viewIndented_rhs_3

  view : Tag type -> String
  view = viewIndented 0

  render : Tag "html" -> String
  render x =
    "<!DOCTYPE HTML>\n" ++ viewIndented 0 x
